/***************************************************************************
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 ***************************************************************************/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <linux/limits.h>

#define VERSION "0.2.1"
#define CONTACT "devel@botland.ath.cx"
#define WRAP "{}\0"

#define CMD_MAX 100
#define LINE_MAX 1000

void version (void)	{
printf ("clwrap %s - %s\n\n", VERSION, CONTACT);
printf ("This program is free software.\n\
You can redistribute it and/or modify it under the terms of the GNU\n\
General Public License as published by the Free Software Foundation;\n\
either version 2 of the License, or (at your option) any later version.\n\
There is NO warranty; not even for MERCHANTABILITY or FITNESS FOR A\n\
PARTICULAR PURPOSE.\n\n");
}

void usage (void)	{
printf ("Usage: clwrap [options]... -e [command]...\n");
printf ("Type `clwrap --help` for more information.\n");
}

void detail (void)	{
printf ("Usage: clwrap [options]... -e [command]...\n");
printf ("Execute a list of commands for each lines in input\n");
printf ("    -e, --execute COMMAND    command to execute where '{}' is replaced\n");
printf ("                             by the input line.\n");
printf ("    -d, --donothing          do nothing.\n");
printf ("    -f, --file FILE          read input file for commands.\n");
printf ("    -h, --help               this help.\n");
printf ("    -i, --interactive        ask confirmation before executing the command.\n");
printf ("    -l, --log FILE           log error to file.\n");
//printf ("    -s, --shell              not yet implemented.\n");
printf ("    -v, --verbose            print each command executed.\n");
printf ("    -V, --version            print version information and exit.\n");
printf ("\nNotes:\n \
   - clwrap require an absolute path if you want to process files outside the current directory.\n \
   - clwrap do not know alias, please use the default shell commands.\n \
\nExamples:\n \
    locate myfiles | clwrap -e cp {} /mydirectory\n \
    locate myfiles | clwrap -e ls -als {}\n \
    lsd | clwrap -e {}/configure -e {}/make -e {}/make install\n \
    alias lsd='find -type d -maxdepth 1 -name \"*\" -printf \"\%d\\n\"' ;-)\n");
}

void outOfLimit (void)	{
printf ("clwrap: too many commands\n");
printf ("Try `clwrap -f FILE` for more commands.\n");
}

void shwstr (char *buffer)     {       printf ("%s\n", buffer); }
void shwint (int buffer)       {       printf ("%d\n", buffer); }

short cmp2str (char *ref, char *cmp1, char *cmp2)	{
	short ret = 0;
	if (!strcmp (ref, cmp1) || !strcmp (ref, cmp2))	ret = 1;
	return ret;
}

int str_char_index (char *str, char c)	{
	int i=0;
	while (i < strlen (str))	{
		if (str[i] == c)	return i;
		i++;
	}
	return -1;
}

int str_is_quoted (char *str)	{
	int i=0;
	if (str[0] == '\"')	i++;
	if (str[strlen (str)] == '\"')	i++;
	return i;
}

/*
 * quote space delimited substring if 'sub' string is in
 */
int substr_quote (char *str, char *sub)	{
	int i=0;
	char res[NAME_MAX];
	char *ptmp;
	char *pres=res;
	char *pstr=str;
	char *psub=sub;
	while (*pstr != '\0' && *pstr != '\n')	{
		if (*pstr == ' ')	{
			ptmp = pstr;
			ptmp++;
			while (*psub != '\0' && *ptmp != ' ' && *ptmp != '\n' && *ptmp != '\0')	{
				while (*ptmp == *psub && *psub != '\0')	{
					psub++;
					ptmp++;
				}
				if (*psub != '\0')	{
					psub = sub;
					ptmp++;
				}
			}
			if (*psub == '\0')	{
				* (pres++) = * (pstr++);
				* (pres++) = '"';
				while (*pstr != ' ' && *pstr != '\n' && *pstr != '\0')	* (pres++) = * (pstr++);
				* (pres++) = '"';
				psub = sub;
				i++;
			}
		}
		* (pres++) = * (pstr++);
	}
	*pres = '\0';
	pres = res;
	pstr = str;
	while (*pres != '\0')	* (pstr++) = * (pres++);
	*pstr = '\0';
	return i;
}

/*
 * change 'old' string in 'ref' string  by 'new' string in 'res' string.
 */
int substr_replace
 (char *const ref, char *const old, char *const new, char *const res, size_t resSize)	{
	int i=0;
	char *pnew = new;
	char *pold = old;
	char *pref = ref;
	char *pres = res;
	char *ptmp;
	while (*pref != '\0' && (pres - res) < resSize)	{
	    	if (*pref == *pold)	{
        		ptmp = pref;
			while (*ptmp == *pold && *pold != '\0' && *ptmp != '\0')	{
				ptmp++;
				pold++;
			}
			if (*pold == '\0')	{
				while (*pnew != '\n' && *pnew != '\0')	* (pres++) = * (pnew++);
 				pnew = new;
				pold = old;
				while (* (pold++) != '\0')	pref++;
				i++;
			}
                        pold = old;
		}
		* (pres++) = * (pref++);
	}
	*pres = '\0';
	if ( (pres - res) > resSize)	i=-1;
	return i;
}

int bld_comm (char *in, char *comm, char *out)	{
	return substr_replace (comm, WRAP, in, out, NAME_MAX);
}

void bld_comm_proto (char *comm)	{
 	substr_quote (comm, WRAP);
}
	
void bld_comm_proto_from_args (int argc, char *argv[], int *iarg, char *comm)	{
	strcpy (comm, argv[*iarg]);
	if (str_char_index (comm, ' ') < 0 && (*iarg + 1) < argc && !cmp2str (argv[ (*iarg) + 1], "-e", "--execute"))	{
		while (*iarg < argc - 1)	{
			strcat (comm, " ");
			strcat (comm, argv[++ (*iarg)]);
		}
	}
	bld_comm_proto (comm);
}

void bld_comm_proto_from_file (char *infile, char comm[CMD_MAX][NAME_MAX], int *icomm)	{
	char *line;
	FILE *file;
	if ( (file = fopen (infile, "r")) != NULL)	{
		while (fgets (line, LINE_MAX, file))	{
			bld_comm_proto (line);
			strcpy (comm[ (*icomm)++], line);
		}
	}
}

int read_pipe (pid_t handle, char *const buffer)	{
	int state, length;
	state = read (handle, &length, sizeof (length));
	if (state > 0)	read (handle, buffer, length);
	else  state = -1;
	return state ;
}

void write_pipe (pid_t handle, char const *const buffer)	{
	int length;
	length = strlen (buffer) + 1;
	write (handle, &length, sizeof (length));
	write (handle, buffer, length);
}

int chkarg
 (int argc, char *argv[], char comm[CMD_MAX][NAME_MAX], char *logf, short *donot, short *inte, short *verb)	{
	int c=0, i;
	if (argc < 2)	i=-1;
	else	i=1;
	while (0 < i && i < argc)	{
		if (argv[i][0] == '-')	{
			if (cmp2str (argv[i], "-e", "--execute"))	{
				if (++i == argc)			i=-1;
				else	{
					bld_comm_proto_from_args (argc, argv, &i, comm[c++]);
					i++;
				}
			}
			else if (cmp2str (argv[i], "-d", "--donothing"))	{
				*donot=1;
				i++;
			}
			else if (cmp2str (argv[i], "-f", "--file"))	{
				if (++i == argc)			i=-1;
				else if (argv[i][0] == '-')		i=-1;
				else	bld_comm_proto_from_file (argv[i++], comm, &c);
			}
			else if (cmp2str (argv[i], "-h", "--help"))	{
				detail ();
				i=0;
			}
			else if (cmp2str (argv[i], "-i", "--interactive")){
				*inte=1;
				i++;
			}
			else if (cmp2str (argv[i], "-l", "--log"))	{
				if (++i == argc)			i=-1;
				else if (argv[i][0] == '-')		i=-1;
				else	strcpy (logf, argv[i++]);
			}
			else if (cmp2str (argv[i], "-v", "--verbose"))	{
				*verb=1;
				i++;
			}
			else if (cmp2str (argv[i], "-V", "--version"))	{
				version ();
				i=0;
			}
			else	i=-1;
		}
		else	i=-1;
	}
//	if (i == argc)	i=-1;
	if (i == -1)	{
		usage ();
		i=0;
	}
	if (c > CMD_MAX)	c=0;
	return c;
}

int main (int argc, char *argv[])	{
	int c, cc;
	int keyb;
	int i, ii=0;
	short donot=0, interact=0, verbose=0;
	char command[CMD_MAX][NAME_MAX];
	char in[NAME_MAX], out[NAME_MAX];		//	buffers
	char logfile[NAME_MAX];				//	file
	char line[NAME_MAX][LINE_MAX];
	in[0] = out[0] = logfile[0] = '\0';
	for (c = 0; c < CMD_MAX; c++)	command[c][0] = '\0'; 
    	cc = chkarg (argc, argv, command, logfile, &donot, &interact, &verbose);
	if (cc > 0)	{
		for (c = 0; c < cc; c++)	{
			if (strlen (logfile) > 0)	{
				strcat (command[c], " 2 >> ");
				strcat (command[c], logfile);
			}
		}
		if (interact)	{
			while (fgets (in, sizeof in, stdin) && ii < LINE_MAX)	strcpy (line[ii++], in);
//			freopen ("/dev/tty", "r", stdin);
			for (i = 0; i < ii; i++)	{
				for (c = 0; c < cc; c++)	{
					bld_comm (line[i], command[c], out);
					keyb = 'w';
					while (keyb != 'y' && keyb != 'n')	{
						printf ("%s ? (y/n) ", out);
						keyb = fgetc (stdin);
						fgetc (stdin);
						if (keyb == 'y')  system (out);
					}
				}
			}
		}
		else	{
			while (fgets (in, sizeof in, stdin))	{
				for (c = 0; c < cc; c++)   {
					bld_comm (in, command[c], out);
					if (verbose)	shwstr (out);
					if (!donot)	system (out);
				}
			}
		}
	}
	return EXIT_SUCCESS;
}
